#include <algorithm>
#include <deque>

#include <windows.h>

#include "conf.hpp"
#include "gfx.hpp"
#include "resource.hpp"
#include "level.hpp"
#include "global.hpp"

#include "data.hpp"
#include "parser.hpp"
#include "machine.hpp"
#include "about_box.hpp"
#include "bug_box.hpp"

LRESULT CALLBACK main_wnd_proc(HWND, UINT, WPARAM, LPARAM);

// The virtual machine executing programs (generating instructions)
machine cpu;

// The output buffer used to store pending instuctions generated by cpu
// and that still await to be executed by the robot
std::deque<char> instruction_buffer;

// Used to temporarily hold program text
char program_buffer[1000];

// Pointer inside program_buffer to the current char during
// analysis
char const* program_buffer_it;

// Id used to the Timer pacing the execution
int const timer_id=100;

// Start running the program
void start_running(HWND hwnd)
{
    SetTimer(hwnd, timer_id, 100, NULL);
    SetWindowText(hwnd, "uHerbert (running)");
}

// Stop running the program
void stop_running(HWND hwnd)
{
    KillTimer(hwnd, timer_id);
    SetWindowText(hwnd, "uHerbert");
}

int lexer()
{
    if(program_buffer_it==program_buffer+sizeof(program_buffer))
        return EOF;
    char c=*program_buffer_it;

    if(c=='\r'){
        ++program_buffer_it;
        if(program_buffer_it==program_buffer+sizeof(program_buffer)) return EOF;
        c=*program_buffer_it;
    }
    program_buffer_it++;
    //printf("Returning '%c'\n", c);
    return c;
}

// declared in global.hpp
HINSTANCE global::hinstance;

// The edit where code is entered
HWND edit;

// The currently loaded and shown level
level current_level;

// Dynamic info about the current level
dyn_level current_dyn_level;

// Invlidate the playground area
void invalidate_playground(HWND hwnd)
{
    RECT dirt;
    dirt.left=0; dirt.top=0;
    dirt.right=conf::playground_width;
    dirt.bottom=conf::playground_height;
    InvalidateRect(hwnd, &dirt, false);
}

// Primitive to move the robot straight
void robot_straight()
{
    int new_x=current_dyn_level.robot_x;
    int new_y=current_dyn_level.robot_y;
    int const old_x=current_dyn_level.robot_x;
    int const old_y=current_dyn_level.robot_y;

    switch(current_dyn_level.robot_dir){
        case direction::up: new_y--; break;
        case direction::down: new_y++; break;
        case direction::left: new_x--; break;
        case direction::right: new_x++; break;
        default: break;
    }

    if(new_y<0 || new_y>=conf::playground_unit_count_y) return;
    if(new_x<0 || new_x>=conf::playground_unit_count_x) return;

    switch(current_level.cells[new_x][new_y]){
        case level::cell::wall:
            return; // Moves in a wall
        case level::cell::white:
            // Press a button
            current_dyn_level.cells[new_x][new_y]|=dyn_level::cell_info::pressed;
            break;
        case level::cell::gray:
            current_dyn_level.unpress_all();
            break;
        case level::cell::dot:
            break;
    }

    switch(current_dyn_level.robot_dir){
        case direction::down:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_down;
        break;
        case direction::up:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_up;
        break;
        case direction::left:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_left;
        break;
        case direction::right:
        current_dyn_level.cells[old_x][old_y] |= dyn_level::cell_info::trail_right;
        break;
    }
    current_dyn_level.robot_x=new_x;
    current_dyn_level.robot_y=new_y;

}

// Primitive to move the robot left (the 'l' instruction)
void robot_left()
{
    switch(current_dyn_level.robot_dir){
        case direction::left:
            current_dyn_level.robot_dir=direction::down; break;
        case direction::right:
            current_dyn_level.robot_dir=direction::up; break;
        case direction::up:
            current_dyn_level.robot_dir=direction::left; break;
        case direction::down:
            current_dyn_level.robot_dir=direction::right; break;
        default: break;
    }
}

// Primitive to move the robot right (the 'r' instruction)
void robot_right()
{
    switch(current_dyn_level.robot_dir){
        case direction::left:
            current_dyn_level.robot_dir=direction::up; break;
        case direction::right:
            current_dyn_level.robot_dir=direction::down; break;
        case direction::up:
            current_dyn_level.robot_dir=direction::right; break;
        case direction::down:
            current_dyn_level.robot_dir=direction::left; break;
        default: break;
    }
}

// Draw static level structures (buttons & walls)
void draw_level(HDC hdc, level const& lev)
{
    draw_grid(hdc);
    for(int y=0; y<conf::playground_unit_count_y; ++y)
    for(int x=0; x<conf::playground_unit_count_x; ++x){
        switch(lev.cells[x][y]){
            case level::cell::white:
                draw_button(hdc, button::white, x, y); break;
            case level::cell::gray:
                draw_button(hdc, button::gray, x, y); break;
            case level::cell::wall:
                draw_wall(hdc, x, y); break;
            default: continue;
        }// switch(lev.cells
    }
}

// Draw dynamic level structures (pressed buttons & trails)
void draw_dyn_level(HDC hdc, dyn_level const& dlev)
{
    for(int y=0; y<conf::playground_unit_count_y; ++y)
    for(int x=0; x<conf::playground_unit_count_x; ++x){
        if(dlev.cells[x][y]&dyn_level::cell_info::pressed)
            draw_button(hdc, button::white_pressed, x, y);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_up)
            draw_trail(hdc, x,y,direction::up);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_down)
            draw_trail(hdc, x,y,direction::down);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_left)
            draw_trail(hdc, x,y,direction::left);
        if(dlev.cells[x][y]&dyn_level::cell_info::trail_right)
            draw_trail(hdc, x,y,direction::right);
    }
    draw_robot(hdc, dlev.robot_dir, dlev.robot_x, dlev.robot_y);

}


int WINAPI WinMain(HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    PSTR cmd_line,
    int cmd_show)
{
    global::hinstance=hInstance;

    char const main_wnd_class_name[]="MAIN_WND_CLASS";

    // Register main window's class
    WNDCLASS wnd_class;
    wnd_class.style=CS_HREDRAW|CS_VREDRAW;
    wnd_class.lpfnWndProc=main_wnd_proc;
    wnd_class.cbClsExtra=wnd_class.cbWndExtra=0;
    wnd_class.hInstance=hInstance;
    wnd_class.hIcon=LoadIcon(hInstance, "HERBERT_ICON");
    wnd_class.hCursor=LoadCursor(NULL, IDC_HAND);
    wnd_class.lpszMenuName="MAINMENU";
    wnd_class.hbrBackground=(HBRUSH)GetStockObject(WHITE_BRUSH);
    wnd_class.lpszClassName=main_wnd_class_name;

    if(! RegisterClass(&wnd_class)){
        MessageBox(NULL, "Couldn't register window class",
            "Fatal Error", MB_ICONERROR);
        return 1;
    }

    // Creating the main window
    HWND hwnd=CreateWindow(main_wnd_class_name,
        "uHerbert",
        WS_SYSMENU|WS_MINIMIZEBOX,
        CW_USEDEFAULT, CW_USEDEFAULT, conf::playground_width,
            conf::playground_height,
        NULL, //parent
        NULL, // menu handle
        hInstance,
        NULL);

    if(! hwnd){
        MessageBox(hwnd, "Couldn't create window", "Fatal Error",
            MB_ICONERROR);
        return 1;
    }

    ShowWindow(hwnd, cmd_show);
    UpdateWindow(hwnd);

    // Program main loop
    MSG msg;
    while(GetMessage(&msg, NULL, 0, 0)){
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    return msg.wParam;
}

LRESULT CALLBACK main_wnd_proc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{

    switch(msg){
        case WM_CREATE:
        {
            // Try to obtain a client area of
            //  conf::playground_width X conf::playground_height
            RECT client; GetClientRect(hwnd, &client);
            const int client_w=client.right;
            const int client_h=client.bottom;

            CREATESTRUCT& cs=*(CREATESTRUCT*)lparam;

            MoveWindow(hwnd, cs.x, cs.y, cs.cx-client_w+conf::playground_width,
                cs.cy-client_h+conf::playground_height+
                conf::edit_height+conf::separator_size, true);

            edit=CreateWindow("EDIT", "(Your program here)",
                WS_CHILD|WS_VISIBLE|ES_MULTILINE|WS_BORDER
                | ES_AUTOHSCROLL| ES_AUTOVSCROLL | WS_VSCROLL,
                0, conf::playground_height+conf::separator_size,
                conf::playground_width,
                conf::edit_height,
                hwnd,
                NULL,
                global::hinstance,
                NULL);

            SendMessage(edit, WM_SETFONT, (WPARAM)GetStockObject(ANSI_FIXED_FONT),
                true);

            return 0;
        }
        case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC hdc=BeginPaint(hwnd, &ps);

            HDC back_dc=CreateCompatibleDC(hdc);
            HBITMAP back_bit=CreateCompatibleBitmap(hdc,
                conf::playground_width, conf::playground_height);
            DeleteObject(SelectObject(back_dc, back_bit));

            draw_level(back_dc, current_level);
            draw_dyn_level(back_dc, current_dyn_level);

            BitBlt(hdc, 0, 0, conf::playground_width, conf::playground_height,
                back_dc, 0, 0, SRCCOPY);

            DeleteDC(back_dc);
            DeleteObject(back_bit);

            EndPaint(hwnd, &ps);
            return 0;
        }// case WM_PAINT

        case WM_CLOSE:
        {
            PostQuitMessage(0);
            return 0;
        }

        case WM_COMMAND:
        {
            switch(LOWORD(wparam)){
                case IDM_LEVEL_OPEN:
                {
                    stop_running(hwnd);

                    char buffer[256]; buffer[0]=0;

                    OPENFILENAME ofn;
                    std::fill((char*)&ofn, (char*)&ofn+sizeof(ofn), 0);

                    ofn.lStructSize=sizeof(ofn);
                    ofn.hwndOwner=hwnd;
                    ofn.hInstance=global::hinstance;
                    ofn.lpstrFilter="Text files (*.txt)\0*.txt\0All files (*.*)\0*.*\0";
                    ofn.lpstrFile=buffer;
                    ofn.nMaxFile=sizeof(buffer);

                    if(GetOpenFileName(&ofn)){
                        FILE* fp=fopen(buffer, "rb");
                        if(! fp){
                            MessageBox(hwnd, "Error opening", buffer, MB_ICONERROR);
                            return 0;
                        }
                        if(!read(current_level, fp)){
                            MessageBox(hwnd, "Parse Error", buffer, MB_ICONERROR);
                        }
                        current_dyn_level.reset(current_level.robot_dir,
                                current_level.robot_x,
                                current_level.robot_y);
                        invalidate_playground(hwnd);
                    }

                    return 0;
                }// case IDM_LEVEL_OPEN
                case IDM_LEVEL_RESET:
                    stop_running(hwnd);
                    current_dyn_level.reset(current_level.robot_dir,
                        current_level.robot_x, current_level.robot_y);
                    invalidate_playground(hwnd);
                    return 0;

                case IDM_LEVEL_CLEAR:
                    stop_running(hwnd);
                    current_level.reset();
                    current_dyn_level.reset();
                    invalidate_playground(hwnd);

                    return 0;

                case IDM_EXEC_START:
                {
                    stop_running(hwnd);

                    GetWindowText(edit, program_buffer, sizeof(program_buffer));
                    program_buffer_it=program_buffer;

                    if(strlen(program_buffer)!=sizeof(program_buffer)-1 &&
                        program_buffer[strlen(program_buffer)-1]!='\n'){
                        strcat(program_buffer, "\n");
                    }

                    parser::get_char=lexer;
                    std::list<function> pgm;
                    if(! parser::yyparse(&pgm)) {
                        MessageBox(hwnd, "Parse Error", "Syntax Error", MB_ICONERROR);
                        return 0;
                    }
                    cpu.load(pgm);
                    instruction_buffer.clear();
                    start_running(hwnd);

                    return 0;
                }
                case IDM_EXEC_STOP:
                    stop_running(hwnd);
                    return 0;
                case IDM_HELP_ABOUT:
                    about_box::show(hwnd);
                    return 0;
                case IDM_HELP_BUGS:
                    bug_box::show(hwnd);
                    return 0;

            }
            return 0;
        } // case WM_COMMAND
        case WM_TIMER:
        {
            while(instruction_buffer.empty()){
                std::vector<char> res;
                try {
                    if(!cpu.exec(res)){
                        stop_running(hwnd);
                        return 0;
                    }
                } catch(char const* str){
                    MessageBox(hwnd, str, "Runtime Error", MB_ICONERROR);
                    stop_running(hwnd);
                    return 0;
                } catch(...) {
                    MessageBox(hwnd, "Unexpected exception", "Runtime Error",
                        MB_ICONERROR);
                    stop_running(hwnd);
                    return 0;
                }
                instruction_buffer.assign(res.begin(),
                    res.end());
            }
            switch(instruction_buffer.front()){
                case 's': robot_straight(); break;
                case 'l': robot_left(); break;
                case 'r': robot_right(); break;
                default: break;
            }
            instruction_buffer.pop_front();
            invalidate_playground(hwnd);
            return 0;
        } // WM_TIMER
    }
    return DefWindowProc(hwnd, msg, wparam, lparam);
}
